;###############################################################################
/* Grid for LilBuilder
06-02-2007(ahklerner)-Added Grid_Draw()

*/
;###############################################################################

;Returns/increases/decreases grid size
GRID(GuiHWND, Action=""){
  static GRID_Increment := 5                         ;initial grid size and increments
  GridSize := Gui(GuiHWND, "GridSize")
  If !Action                                         ;return grid size when nothing is specified
      Return GridSize

  GridSize := GridSize ? GridSize : "0"

  If Action = Up                                     ;increase GRID on "UP"
    GridSize += GRID_Increment
  Else If Action = Down                              ;increase GRID on "Down"
    GridSize -= GRID_Increment
  Else If Action is integer
    GridSize = %Action%
  Else                                               ;issue error otherwise
    MsgBox, Wrong Action in Grid()

  GridSize := GridSize < 0 ? 0 : GridSize            ;make sure it is >= 0
  ToolTip("Grid = " GridSize, 2000, "Continue")      ;give feedback
  Gui(GuiHWND, "GridSize", GridSize)
  Return GridSize                                    ;return new GRID value
}

;change grid size
;called when
; - hotkeys to change grid size are pressed 
Grid_Change:
  Instr(A_ThisHotkey, "F6") ? GRID(Selector_vGuiHWND(),"Up") : GRID(Selector_vGuiHWND(),"Down")   ;increase/decrease pending on hotkey
  Grid_Draw(Selector_vGuiHWND())                             ;update grid
Return

;draw the grid
Grid_Draw(GuiHWND, $Color="0xC0C0C0",$Weight=1) {
  Static Grid_Old$Size, cnt = 0
  
  If !($Size := GRID(GuiHWND)){                ;return, when no grid is used
      If Grid_Old$Size {                            ;redraw gui, when it has been grid before
         WinSet, Redraw,, ahk_id %GuiHWND%
         Grid_Old$Size = 0
      }
      Return
  }

  ; Get rid of all of the old lines
  If (Grid_Old$Size != $Size) {
    WinSet, Redraw,, ahk_id %GuiHWND%
    Sleep, 50      ; Wait a bit...otherwise some lines are not drawn
  }
  Grid_Old$Size := $Size

  ;Get the position of the builder window
  WinGetPos, Grid_Line_x, Grid_Line_y, Grid_Window_h, Grid_Window_w, ahk_id %GuiHWND%

  ;figure out how many lines to draw
  Grid_Line_Count_W := Grid_Window_w / $Size, Grid_Line_Count_H := Grid_Window_h / $Size

  ; get the device context of the builder window
  Grid_hDC := DllCall("GetDC", UInt, GuiHWND)

  ; Create a pen to draw with
  Grid_hCurrPen := DllCall("CreatePen", UInt, 0, UInt, $Weight, UInt, $Color)
  DllCall("SelectObject", UInt,Grid_hDC, UInt,Grid_hCurrPen)


  ; Draw the grid lines
  Grid_Line_x = 0
  Loop, % Grid_Line_Count_H
  {
    ; increase the x position by the Grid size
    Grid_Line_x += $Size
    ; Set the begining of the line
    DllCall("gdi32.dll\MoveToEx", UInt, Grid_hDC, Uint,Grid_Line_x, Uint, 0, Uint, 0 )
    ; Set the end of the line
    DllCall("gdi32.dll\LineTo", UInt, Grid_hDC, Uint, Grid_Line_x, Uint, Grid_Window_w)
  }
  
  Grid_Line_y = 0
  Loop, % Grid_Line_Count_W
  {
    ; increase the y position by the Grid size
    Grid_Line_y += $Size
    ; Set the beginning of the line
    DllCall("gdi32.dll\MoveToEx", UInt, Grid_hDC, Uint,0, Uint, Grid_Line_y, Uint, 0 )
    ; Set the end of the line
    DllCall("gdi32.dll\LineTo", UInt, Grid_hDC, Uint, Grid_Window_h, Uint, Grid_Line_y)
  }

  ; Cleanup
  DllCall("ReleaseDC", UInt, 0, UInt, Grid_hDC)  ; Clean-up.
  DllCall("DeleteObject", UInt,Grid_hCurrPen)
}
