;create complete gui (one or all in project) with all its controls
LB_BuildGui(GuiIDtoCreate = 0){
  If !GuiIDtoCreate {
     ListOfGuiID := LB_GetPropValue("Project", "ListOfGuiID", A_Space)
     Loop, Parse, ListOfGuiID, |, A_Space
       LB_CreateGui(A_LoopField)
  }Else
    LB_CreateGui(GuiIDtoCreate)
}

;create a gui
LB_CreateGui(GuiID){
  ;get all gui properties
  AllGuiProps := LB_GetPropValue("Gui " GuiID)
  Loop, Parse, AllGuiProps, `n
  {
    Pos := InStr(A_LoopField,"=")
    StringLeft, Prop, A_LoopField, Pos - 1
    StringTrimLeft, Value, A_LoopField, Pos
    %Prop% := Value
  }

  ;collect gui options
  Options := (Toolwindow ? " +Toolwindow" : "")
           . (Resize ? " +Resize" : "")
           . (Label ? " +Label" Label : "")
           . (AlwaysOnTop ? " +AlwaysOnTop" : "")

  ;apply gui options
  SaveLoad_RealScript("Gui, " GuiID ": +LastFound " Options
                     ,"Gui, %GuiID%: +LastFound %Options%")
  
  ;make guis HWND global
  LB_SetVarGlobal("Gui" GuiID "_HWND", WinExist())
  SaveLoad_RealScript("Gui" GuiID "_HWND = WinExist()"
                     ,"")

  ;create controls
  TabCount = 0
  Loop, Parse, OrderOfControls, |, A_Space
      LB_CreateControl(GuiID, A_LoopField, TabCount)

  ;collect gui show options
  Options := (Hidden ? " Hide" : "")

  ;show gui
  SaveLoad_RealScript("Gui, " GuiID ":Show, " PosSize " " Options ", " Content
                     ,"Gui, %GuiID%:Show, %PosSize% %Options%, %Content%")
}


;set a specific var global
LB_SetVarGlobal(VarName, VarValue){
  global
  %VarName% = %VarValue%
}

;create a control
LB_CreateControl(GuiID, CtrlNumber, ByRef TabCount){
  ;get all control properties
  AllCtrlProps := LB_GetPropValue("Control " CtrlNumber)
  Loop, Parse, AllCtrlProps, `n
  {
    Pos := InStr(A_LoopField,"=")
    StringLeft, Prop, A_LoopField, Pos - 1
    StringTrimLeft, Value, A_LoopField, Pos
    %Prop% := Value
  }

  ;collect control options
  Options := PosSize " " Options
          . (VarName ? " v" VarName : "")
          . (Label ? " g" Label : "")
          . (Disabled ? " Disabled" : "")
          . (Checked ? " Checked" : "")
          . (Hidden ? " Hidden" : "")

  ;add control to gui in global context
  LB_CreateControlAdd(GuiID, Type, Options, Content)

  ;collect and apply control options, that can only be applied after creation

  ;recurse into tab controls
  If (Type = "Tab"){
    ThisTabCnt := ++TabCount      ;increase tab count and memorize it for recursion
    Loop, Parse, OrderOfTabPageControls, |, A_Space    ;go through all tab pages of tab control
    {
      ThisTabPage := A_Index
      ;activate current tab page of current tab control for its controls
      SaveLoad_RealScript("Gui, " GuiID ":Tab, " ThisTabPage ", " ThisTabCnt
                         ,"Gui, %GuiID%:Tab, %ThisTabPage%, %ThisTabCnt%")
      OrderOfControls := LB_GetPropValue("TabPage " A_LoopField, "OrderOfControls", A_Space)
      Loop, Parse, OrderOfControls, |, A_Space         ;go through all controls of tab page
         If ("Tab" = LB_CreateControl(GuiID, A_LoopField, TabCount))    ;create the control
         { ;when control on tab page was a tab control, re-activate current tab page for new controls
           SaveLoad_RealScript("Gui, " GuiID ":Tab, " ThisTabPage ", " ThisTabCnt
                              ,"Gui, %GuiID%:Tab, %ThisTabPage%, %ThisTabCnt%")
         }
    }
    ;close tab control for new controls
    SaveLoad_RealScript("Gui, " GuiID ":Tab"
                       ,"Gui, %GuiID%:Tab")

  }
  Return Type
}

;add control to gui in global context
LB_CreateControlAdd(GuiID, Type, Options, Content){
  global
  SaveLoad_RealScript("Gui, " GuiID ":Add, " Type ", " Options ", " Content
                     ,"Gui, %GuiID%:Add, %Type%, %Options%, %Content%")
}

;anchor controls on gui
LB_Anchor(c, a = "", r = false) { ; v3.6 - Titan
	static d
	GuiControlGet, p, Pos, %c%
	If ex := ErrorLevel {
		Gui, %A_Gui%:+LastFound
		ControlGetPos, px, py, pw, ph, %c%
	}
	If !(A_Gui or px) and a
		Return
	i = x.w.y.h./.7.%A_GuiWidth%.%A_GuiHeight%.`n%A_Gui%:%c%=
	StringSplit, i, i, .
	d := a ? d . ((n := !InStr(d, i9)) ? i9 : "")
		: RegExReplace(d, "\n\d+:" . c . "=[\-\.\d\/]+")
	Loop, 4
		x := A_Index, j := i%x%, i6 += x = 3
		, k := !RegExMatch(a, j . "([\d.]+)", v) + (v1 ? v1 : 0)
		, e := p%j% - i%i6% * k, d .= n ? e . i5 : ""
		, RegExMatch(d, "\Q" . i9 . "\E(?:([\d.\-]+)/){" . x . "}", v)
		, l .= p%j% := InStr(a, j) ? (ex ? "" : j) . v1 + i%i6% * k : ""
	If r
		rx = Draw
	If ex
		ControlMove, %c%, px, py, pw, ph
	Else GuiControl, Move%rx%, %c%, %l%
}

;return the data requested from the project data
LB_GetPropValue(Sect, Key = "", Default = "", ProjectData = ""){
  If !ProjectData
      ProjectData := LB_DummyProjectData()

  ;if a key is requested, return its value or when not found return the default value
  If Key {
    re = miUS)\Q[%Sect%]\E.*\R\Q%Key%\E=(?P<Value>.*)(\R|$)
    Return RegExMatch(ProjectData, re, Key) ? KeyValue : Default
  }

  ;otherwise return all keys with their values as a text block
  re = sU)\Q[%Sect%]\E\R(?P<Value>.*)(\R\[|$)
  RegExMatch(ProjectData, re, Key)
  Return KeyValue
}

LB_DummyProjectData(ProjectData = ""){
  ;dummy function, that allows to include this function code into LilBuilder directly.
  ;its function call will be replaced with LB_ProjectData() on export
  ;all code below will get deleted on export
  Static Data
  If ProjectData
      Data = %ProjectData%
  Return Data
}
