;###############################################################################
/* SaveLoad for LilBuilder

*/
;###############################################################################

SaveLoad_ExportProject(OutFile, PreFix, ChkRealAHKScript, ChkRunExport, ChkExportOnlyData
                        , ChkExportOnlyDataAndFunc){

  ;only add extention ahk if it is a full or real AHK script
  If (ChkRealAHKScript OR ChkRunExport)
    OutFile .= RegExMatch(OutFile, "i)\.ahk$") ? "" : ".ahk"

  ;delete old version of OutFile
  FileDelete, %OutFile%

  ;store export settings
  Project("ProjectName", ProjectName)
  Project("RunExport", ChkRunExport)
	Project("ExportOnlyData", ChkExportOnlyData)
	Project("ExportOnlyDataAndFunc", ChkExportOnlyDataAndFunc)
	Project("_LastSession", OutFile)

  ;collect different parts of final ouput
  SaveLoad_CreateFinalScript(PreFix, AutoExecSection, CloseSection, AnchorSection, LabelSection, LabelNameSection, VarNameSection)
  FunctionsCode := SaveLoad_ProjectFunctions("Save", PreFix)
  ProjectData := SaveLoad_ProjectData("Save", PreFix)
  CodeOfSession := Project("_CodeOfSession")

  ;append different parts to output
  If ChkRealAHKScript {
     ProjectData := SaveLoad_CollectProjectData()
     LB_DummyProjectData(ProjectData)
     LB_BuildGui()
     Script := SaveLoad_RealScript()
     FileAppend, %LabelNameSection%%VarNameSection%`n%Script%`n`n%CodeOfSession% , %OutFile%
  }Else{
    If (ChkExportOnlyData OR ChkExportOnlyDataAndFunc){
      FileAppend,  , %OutFile%
    }Else{
      ;add code that got imported with the project and doesn't belong to LilBuilder Code
      If CodeOfSession
        FileAppend, %CodeOfSession%`n`n%LabelNameSection%%VarNameSection%, %OutFile%
      Else
        FileAppend, %AutoExecSection%%CloseSection%%AnchorSection%%LabelSection%%VarNameSection% , %OutFile%
    }

    ;append project functions
    If !ChkExportOnlyData
      FileAppend, %FunctionsCode%, %OutFile%

    ;append project data function
    FileAppend, %ProjectData%, %OutFile%
  }
  
  If ChkRunExport
	   Run, %OutFile%
}

SaveLoad_RealScript(ScriptLine = "", Dummy = ""){
  Static Script
  If ScriptLine
    Script .= ScriptLine "`n"
  Else{
    RetScript = %Script%
    Script =
  }
  Return RetScript
}

SaveLoad_ProjectFunctions(Action, ByRef PreFix, File = ""){
static StartText = "/* ##### section of Li'l Builder project functions #####"
    ,  EndText = "/* ##### end of Li'l Builder project functions #####"
    ,  Extention = "
(
`n           !!! Don't edit above line !!!
This comment is used to find this section of the script by Li'l Builder

It is also not recommended to modify the functions or data in this section,
since the changes might get overwritten, once you open this script with
Li'l Builder again.
*/`n`n
)"

  If (Action = "Save"){
    ;read functions code
    FileRead, FunctionsCode, inc\LB_IncludeFunctions.ahk

    ;replace dummy function call with real ProjectData function call
    StringReplace, FunctionsCode, FunctionsCode, LB_DummyProjectData(), LB_ProjectData()

    ;remove dummy function
    FunctionsCode := RegExReplace(FunctionsCode, "mis)^\QLB_DummyProjectData(ProjectData = \E.*")

    ;replace SaveLoad_RealScript() function with real code (second parameter)
    re = i)SaveLoad_RealScript\(.*?\R.*?,"(.*?)"\)
    FunctionsCode := RegExReplace(FunctionsCode, re, "$1")

    ;Prefix LB functions with PreFix
    StringReplace, FunctionsCode, FunctionsCode, LB_, %PreFix%LB_, All
    Return "`n" StartText Extention FunctionsCode "`n" EndText Extention
  }
  Else If (Action = "Load"){
    ;find project function section and remove it from file
    re = mis)\Q%StartText%\E.*?\R\Q%EndText%\E.*?\R\Q*/\E
    RegExMatch(File, re, ProjectSection)
    StringReplace, File, File, %ProjectSection%
    Return File
  }
}

SaveLoad_ProjectData(Action, ByRef PreFix, ByRef File = ""){
static StartText = "/* ##### section of Li'l Builder project data #####"
    ,  EndText = "/* ##### end of Li'l Builder project data #####"
    ,  Extention = "
(
`n           !!! Don't edit above line !!!
This comment is used to find this section of the script by Li'l Builder

Only modify the data in this section when you know what you are doing.
!!! Don't edit anything else (code, comments, function and variable definitions) !!!
*/`n`n
)"
, BeginFunction = "
(
LB_ProjectData(){
;Static ProjectDataPartX variables hold the data of the project.
;Each ProjectDataPartX variable can't hold more then 16383 characters!
;Inside the ProjectDataPartX variables the project/gui/control data is stored in an INI format.
Static Projectdata
)"
, EndFunction = "
(
; end of static ProjectDataPartX variables

  ;return project data stored in the static PropX variables
  ;concatenate only once for speed reasons
  If ProjectData
      Return ProjectData
  Return ProjectData := ConcatenationOfPropXVars
} ; end of LB_ProjectData()`n`n
)"

  If (Action = "Save"){

    ProjectData := SaveLoad_CollectProjectData()

    ;Split ProjectData in parts of 16383 characters
    Loop {
      ProjectDataPart := A_Index
      If (StrLen(ProjectData) < 16370){
        ProjectDataPart%ProjectDataPart% = %ProjectData%
        Break
      }
      ;get position of `n backwards from position 16370
      Pos := InStr(SubStr(ProjectData,1,16370), "`n", 0, 0)
      StringLeft, ProjectDataPart%ProjectDataPart%, ProjectData, Pos - 1
      StringTrimLeft, ProjectData, ProjectData, Pos
    }

    ;Prefix LB functions with PreFix
    StringReplace, FinalBeginFunction, BeginFunction, LB_, %PreFix%LB_, All
    StringReplace, FinalEndFunction, EndFunction, LB_, %PreFix%LB_, All

    ;correct the concatenation in the EndFunction text
    Loop %ProjectDataPart%{
      If A_Index > 1
         ConcatenationOfPropXVars .= " ""`n"" "
      ConcatenationOfPropXVars .= "ProjectDataPart" A_Index
    }
    StringReplace, FinalEndFunction, FinalEndFunction, ConcatenationOfPropXVars, %ConcatenationOfPropXVars%, All

    ;combine all parts to form LB_GetPropValue()
    LB_ProjectData := "`n" StartText Extention FinalBeginFunction
    Loop %ProjectDataPart%
      LB_ProjectData .= ", ProjectDataPart" A_Index " = ""`n(`n" ProjectDataPart%A_Index% "`n)"" "
    Return LB_ProjectData FinalEndFunction EndText Extention
  }
  Else If (Action = "Load"){
    ;find project data section and remove it from file
    re = mis)^\Q%StartText%\E.*?^\Q%EndText%\E.*?^\Q*/\E
    RegExMatch(File, re, ProjectSection)
    StringReplace, File, File, %ProjectSection%
    
    ;find prefix if there is any
    re = mi)^(?P<Fix>.*?)\QLB_ProjectData(){\E
    RegExMatch(ProjectSection, re, Pre)

    ;find project data
    re = mis)\Q%PreFix%LB_ProjectData(){\E.*^Static Projectdata,(?P<Data>.*?\R\)") `; end of static ProjectDataPartX variables$
    RegExMatch(ProjectSection, re, Project)
    
    ;clean ProjectData, from ProjectDataPart text
    re = m)[, ]*ProjectDataPart\d+.*?\R.*\R
    ProjectData := RegExReplace(ProjectData, re)
    re = m)\R[ \t]*\)".*
    ProjectData := RegExReplace(ProjectData, re)

    Return ProjectData
  }
}

;collect project data in INI Format and return it
SaveLoad_CollectProjectData(){
  ProjectData := "[Project]"     ;start with project section

  ;store project properties
  ProjPNames := Project("AllPropertyNames")
	Loop, Parse, ProjPNames, `n
	{
		%A_LoopField% := Project(A_LoopField)   ;!!! needs to be stored in %A_LoopField% to allows reference to ListOfGuiHWND later

		If (A_LoopField = "ListOfGuiHWND")       ;prevent data to be stored
      Continue
    If (InStr(A_LoopField,"_") = 1)
      Continue

		ProjectData .= "`n" A_LoopField "=" %A_LoopField%   ;store data
	}

  ;collect data of guis and store in ProjectData
  ProjectData .= "`nListOfGuiID="    ;create empty, to make sure this is the first in ini file
	Loop, Parse, ListOfGuiHWND, |
    ListOfGuiID .= "|" SaveLoad_CollectGuiData(ProjectData, A_LoopField)
	StringTrimLeft, ListOfGuiID, ListOfGuiID, 1
  StringReplace, ProjectData, ProjectData, `nListOfGuiID=, `nListOfGuiID=%ListOfGuiID%
  Return ProjectData
}

SaveLoad_CollectGuiData(ByRef ProjectData, GuiHWND){
  SysGet_GuiFrameSizes(Gui(GuiHWND,"Toolwindow"), XFrame, YFrame, HCaption)
  FinalGuiID := Gui(GuiHWND,"_FinalGuiID")

	ProjectData .= "`n[Gui " FinalGuiID "]"     ;start gui section

	;store gui size and pos info
  WinGetPos, X, Y, W, H, ahk_id %GuiHWND%
	W -= 2 * XFrame
	H -= 2 * YFrame + HCaption
  ProjectData .= "`nPosSize=x" X " y" Y " w" W " h" H
  
  ;store gui properties
  GuiPNames := Gui(GuiHWND,"AllPropertyNames")
  Loop, Parse, GuiPNames, `n
  {
  	%A_LoopField% := Gui(GuiHWND, A_LoopField)   ;!!! needs to be stored in %A_LoopField% to allows reference to OrderOfControls later

    If A_LoopField in OrderOfControls,GuiID      ;prevent data to be stored
      Continue
    If InStr(A_LoopField,"_") = 1
      Continue

		ProjectData .= "`n" A_LoopField "=" %A_LoopField%   ;store data
	}

  ;collect data of controls and store in ProjectData
  ProjectData .= "`nOrderOfControls=Gui" FinalGuiID   ;create empty with identifier, to replace it robustly later
	CtrlCnt = 0
	Loop, Parse, OrderOfControls, |
  {
    ;don't store stausbar that is deleted
    If ("Statusbar" = Control(A_LoopField, "Type") AND !Gui(GuiHWND, "_StatusBar") )
        Continue
        
    ListOfCtrl .= "|" SaveLoad_CollectControlData(ProjectData, GuiHWND, CtrlCnt, A_LoopField, XFrame, YFrame + HCaption)
	}
  StringTrimLeft, ListOfCtrl, ListOfCtrl, 1
  StringReplace, ProjectData, ProjectData, `nOrderOfControls=Gui%FinalGuiID%, `nOrderOfControls=%ListOfCtrl%

  Return FinalGuiID
}

SaveLoad_CollectControlData(ByRef ProjectData, GuiHWND, ByRef CtrlCnt, CtrlHWND, XOffset, YOffset){
  ThisControlCnt := ++CtrlCnt
  
	ProjectData .= "`n[Control " ThisControlCnt "]"     ;start control section

  ;store control properties
	CtrlPNames := Control(CtrlHWND,"AllPropertyNames")
	Loop, Parse, CtrlPNames, `n
	{
		%A_LoopField% := Control(CtrlHWND, A_LoopField)    ;!!! needs to be stored in %A_LoopField% to allows reference to _FinalVarName,Type,etc. later

    If A_LoopField in BelongsTo,VarName                ;prevent data to be stored
      Continue
    If InStr(A_LoopField,"_") = 1
      Continue
    If (A_LoopField = "AnchorValue" AND !InStr("|" Gui(GuiHWND, "_AnchoredCtrls") "|", "|" CtrlHWND "|"))
      Continue

		ProjectData .= "`n" A_LoopField "=" %A_LoopField%   ;store data
	}

  ;store data with a different name
  If _FinalVarName
      ProjectData .= "`nVarName=" _FinalVarName

  If (Type != "StatusBar"){ 
  	ControlGetPos, X, Y, W, H,, ahk_id %CtrlHWND%
  	X -= XOffset
  	Y -= YOffset
  	ProjectData .= "`nPosSize=x" X " y" Y " w" W " h" H
  }

  ;recurse into tab control to export controls on its pages
  If (Type = "Tab"){
    ProjectData .= "`nOrderOfTabPageControls=Tab" ThisControlCnt   ;create empty with identifier, to replace it robustly later
  	Loop, Parse, Content, |
    {
      ThisTabPageControlCnt := ++CtrlCnt        ;tab page gets its own control number
      OrderOfTabPageControls .= "|" CtrlCnt     ;collect tab page numbers

    	ProjectData .= "`n[Tabpage " ThisTabPageControlCnt "]"             ;start tabpage section
      ProjectData .= "`nOrderOfControls=Tabpage" ThisTabPageControlCnt   ;create empty with identifier, to replace it robustly later

      ;collect data of controls on tab page
      OrderOfControls := Control(CtrlHWND "T" A_Index, "OrderOfControls")
    	ListOfCtrl =
      Loop, Parse, OrderOfControls, |
          ListOfCtrl .= "|" SaveLoad_CollectControlData(ProjectData, GuiHWND, CtrlCnt, A_LoopField, XOffset, YOffset)
      StringTrimLeft, ListOfCtrl, ListOfCtrl, 1
      StringReplace, ProjectData, ProjectData, `nOrderOfControls=Tabpage%ThisTabPageControlCnt%, `nOrderOfControls=%ListOfCtrl%
    }
    StringTrimLeft, OrderOfTabPageControls, OrderOfTabPageControls, 1
    StringReplace, ProjectData, ProjectData, `nOrderOfTabPageControls=Tab%ThisControlCnt%, `nOrderOfTabPageControls=%OrderOfTabPageControls%
  }

  Return ThisControlCnt
}

;###############################################################################

;Steal(copy) gui and open it as project
SaveLoad_StealImport(){
	SaveLoad_ImportProject(Stealer())
}

;###############################################################################

;function to import a project
SaveLoad_ImportProject(ProjectData = "", GuitoCreate = 0){
	If !ProjectData	{
		FileSelectFile, File, 1,, Import a Lil' Builder Project File, Lil Builder Project File (*)
		If !FileExist(File)
        Return

    Project("_LastSession", File)

    ;read file into memory
    FileRead, File, %File%

    ;extract the project data and remove it from the file
    ProjectData := SaveLoad_ProjectData("Load", PreFix, File)

    ;remove functions from file
    File := SaveLoad_ProjectFunctions("Load", PreFix, File)

    ;remove info comments from file
    re = mis)\Q/* ##### Controls with Labels in Li'l Builder project #####\E.*?\Q*/\E
    File := RegExReplace(File, re)
    re = mis)\Q/* ##### Controls with VarNames in Li'l Builder project #####\E.*?\Q*/\E
    File := RegExReplace(File, re)

    ;remove leading and trailing empty lines from file
    re = s)^\s*(.*?)\s*$
    File := RegExReplace(File, re,"$1")

    ;store remaining code from file as code of session
    If File {
      CodeOfSession := Project("_CodeOfSession")
      CodeOfSession .= (CodeOfSession ? "`n`n" : "") File
      Project("_CodeOfSession", CodeOfSession)
    }
	}

  ;read project options and react on them
  ;get all gui properties
  AllProjectProps := LB_GetPropValue("Project","","",ProjectData)
  Loop, Parse, AllProjectProps, `n,`r
  {
    Pos := InStr(A_LoopField,"=")
    StringLeft, Prop, A_LoopField, Pos - 1
    StringTrimLeft, Value, A_LoopField, Pos
    If (Prop = "ListOfGuiID")    ;store ListOfGuiIDs for later reference
      %Prop% := Value
    Else If (Prop = "ProjectName"){
      ;Only when project name is the default name, store new project name
      If ("NewProject" = Project("ProjectName"))
        Project(Prop, Value)
    }
    Else
      Project(Prop, Value)  ;store all project values directly in dictionary
  }

  If !GuitoCreate {
     Loop, Parse, ListOfGuiID, |, A_Space
         SaveLoad_ImportGui(ProjectData, A_LoopField)
  }Else
     SaveLoad_ImportGui(ProjectData, GuitoCreate)
  
}

;function to import a gui
SaveLoad_ImportGui(ProjectData, FinalGuiID){
  ;get all gui properties
  AllGuiProps := LB_GetPropValue("Gui " FinalGuiID,"","",ProjectData)
  Loop, Parse, AllGuiProps, `n,`r
  {
    Pos := InStr(A_LoopField,"=")
    StringLeft, Prop, A_LoopField, Pos - 1
    StringTrimLeft, Value, A_LoopField, Pos
    %Prop% := Value
  }

  ;create Gui (with internal default properties, and hidden)
  ShowOptions := PosSize . " Hide"
  GuiHWND := Gui_Create(FinalGuiID, Content, ShowOptions)
  GuiID := Gui(GuiHWND,"GuiID")       ;get internal GuiID

  ;apply and store gui properties (e.g. toolwindow) after creation
  Items2Apply = Toolwindow
  Loop, Parse, Items2Apply, |
  {
    If (%A_LoopField%) {
        Gui, %GuiID%:+%A_LoopField%
        Gui(GuiHWND,A_LoopField,%A_LoopField%)
    }
  }

  ;store gui properties (e.g. Resize) that are not applied
  Items2Store = Resize|AlwaysOnTop|Hidden
  Loop, Parse, Items2Store, |
    Gui(GuiHWND,A_LoopField,%A_LoopField%)

  ;create controls
  Loop, Parse, OrderOfControls, |, A_Space
      SaveLoad_ImportControl(ProjectData, GuiHWND, A_LoopField)

  ;show gui
  Gui, %GuiID%:Show
  
  If GridSize {
    GRID(GuiHWND, GridSize)
    Grid_Draw(GuiHWND)
  }
}

;function to import a control
SaveLoad_ImportControl(ProjectData, GuiHWND, CtrlNumber){
  ;get all control properties
  AllCtrlProps := LB_GetPropValue("Control " CtrlNumber,"","",ProjectData)
  Loop, Parse, AllCtrlProps, `n, `r
  {
    Pos := InStr(A_LoopField,"=")
    StringLeft, Prop, A_LoopField, Pos - 1
    StringTrimLeft, Value, A_LoopField, Pos
    %Prop% := Value
  }

  ;??? apply control options (e.g. Font, Color), that need to be applied before creation
  
  ;create control (with internal default properties)
  AddOptions := PosSize " " Options
             . (Disabled ? " Disabled" : "")
             . (Checked ? " Checked" : "")
  CtrlHWND := Control_New(Type, GuiHWND, Content, AddOptions)

  ;apply control options, that can only be applied after creation and apply them
  If AnchorValue {
    Control(CtrlHWND,"AnchorValue", AnchorValue)
    AnchoredCtrls := Gui(GuiHWND, "_AnchoredCtrls") . "|" . CtrlHWND
    If (InStr(AnchoredCtrls,"|") = 1)
      StringTrimLeft, AnchoredCtrls, AnchoredCtrls, 1
    Gui(GuiHWND, "_AnchoredCtrls", AnchoredCtrls)
  }

  ;store control options, that are applied on creation or that can not be applied
  Control(CtrlHWND,"_FinalVarName",VarName)
  Items2Store = Label|Disabled|Checked|Hidden
  Loop, Parse, Items2Store, |
    Control(CtrlHWND,A_LoopField,%A_LoopField%)

  ;handle tabs
  If (Type = "Tab"){
    Loop, Parse, OrderOfTabPageControls, |, A_Space
    {
      Tab_AddNewControlsToPage(GuiHWND, CtrlHWND, A_Index)
      OrderOfControls := LB_GetPropValue("TabPage " A_LoopField,"OrderOfControls","",ProjectData)
      Loop, Parse, OrderOfControls, |, A_Space
         SaveLoad_ImportControl(ProjectData, GuiHWND, A_LoopField)
    }
    Tab_AddNewControlsToPage(GuiHWND, "")
  }
}

;###############################################################################

SaveLoad_CreateFinalScript(PreFix, ByRef AutoExecSection, ByRef CloseSection, ByRef AnchorSection
                         , ByRef LabelSection, ByRef LabelNameSection, ByRef VarNameSection){
  AutoExecSection =
    (LTrim
      #NoEnv
      SetBatchLines, -1
      ;Example how to build guis: LB_BuildGui([Gui_ID]) - if Gui_ID is not specified, all Guis in project are built
      %PreFix%LB_BuildGui()   ;build all guis with a single function call
      Return`n`n
    )

	ListOfGuiHWND := Project("ListOfGuiHWND")
	Loop, Parse, ListOfGuiHWND, |
	{
    GuiID := Gui(A_LoopField, "_FinalGuiID")

    If !(GuiLabel := Gui(A_LoopField, "Label"))
      GuiLabel := (GuiID != 1 ? GuiID : "") "Gui"

    If (ListOfCtrlWHND := Gui(A_LoopField, "_AnchoredCtrls")){
      GuiID := Gui(A_LoopField, "GuiID")
      AnchorSection .= "`n" GuiLabel "Size:`n"
      Loop, Parse, ListOfCtrlWHND, |, A_Space
        {
          CtrlID := Control(A_LoopField, "_FinalVarName")
          AnchorValue := Control(A_LoopField, "AnchorValue")
          AnchorSection .= A_Tab PreFix "LB_Anchor( """ CtrlID """, """ AnchorValue """)`n"
        }
      AnchorSection .= "Return`n"
    }

    ListOfCtrlWHND := Gui_GetAllCtrlHWND(A_LoopField)
    Loop, Parse, ListOfCtrlWHND, |, A_Space
      {
        If (Label := Control(A_LoopField, "Label"))
            LabelSection .= Label ":`n"
        If (VarName := Control(A_LoopField, "_FinalVarName"))
            VarNameSection .= VarName "`n"
      }

    CloseSection .= GuiLabel "Close:`n"
  }
  CloseSection = `;Example to close guis`n%CloseSection%%A_Tab%ExitApp`nReturn`n`n

  If AnchorSection
    AnchorSection = `;Example to apply Anchors%AnchorSection%`n`n
  If LabelSection {
    LabelNameSection = /* ##### Controls with Labels in Li'l Builder project #####`n%LabelSection%*/`n`n
    LabelSection = `;Example of control subroutines`n%LabelSection%%A_Tab%Msgbox, `%A_Gui`% `%A_GuiControl`%`nReturn`n`n
  }
  If VarNameSection
    VarNameSection = /* ##### Controls with VarNames in Li'l Builder project #####`n%VarNameSection%*/`n`n
}
