;###############################################################################
/* Selector for LilBuilder

*/
;###############################################################################

;create selector gui
Selector_Create() {                     
   Gui, 1:+LastFound -Caption +Resize +ToolWindow +LabelSelector_ ;AlwaysOnTop
   Gui, 1:Color, FFFF
   WinSet, Transparent, 100
   Selector_vHWND(WinExist())
}

;###############################################################################
;##################   Functions for otherwise global vars                                                    
;###############################################################################

Selector_vHWND(NewHWND=""){            ;selectors own gui HWND
  static HWND
  If NewHWND
     HWND = %NewHWND%
  Else Return %HWND%
}

Selector_vGuiHWND(NewHWND=""){        ;currectly "active" (selected) user gui
  global
  static HWND
  If NewHWND = No                       ;there should normally always be one, toolbox relies on it
     HWND =
  Else If NewHWND
     Dock_HostId := HWND := NewHWND
  Else Return %HWND%
}

Selector_vCtrlHWND(NewHWND=""){      ;list of selected controls in selection, can be a single control
  static HWND
  If NewHWND = No
     HWND =
  Else If NewHWND {
     HWND .= "|" NewHWND
     If InStr(HWND,"|") = 1
        StringTrimLeft, HWND, HWND, 1 
  }Else Return %HWND%
}

;store and retrieve position and size of selector (mainly for multi selection)
Selector_vPosSize(ByRef X, ByRef Y, ByRef W, ByRef H){
  static sX, sY, sW, sH
  If X
    sX := X, sY := Y, sW := W, sH := H,
  Else X := sX, Y := sY, W := sW, H := sH,
}

;###############################################################################
;##################   Position selector to something                                                       
;###############################################################################

;set position and size of selector gui
Selector_Set2PosSize(X, Y, W = 6, H = 6){
  w -= 6, h -= 6
  Gui, 1:Show, x%x% y%y% w%w% h%h% NA
}

;set selector to the dimensions of a control
; called when:
;   - selector needs to update itself after selector got resized/moved with GRID in single selection mode
;   - after a new control gets added via toolbox
;   - a single control gets selected with a click
;   - a single control gets selected with frame selection
;   - a single control gets add to multiselection
;   - multiple controls get selected with frame selection
;   - gui resizes (for anchored controls)
Selector_Set2Ctrl(){
   WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()           ;get pos of user gui
   If InStr(CtrlHWND := Selector_vCtrlHWND(),"|"){
      maxx := 0 ,            maxy := 0                   ;set extreme start coordinates for surrounding box
      minx := A_ScreenWidth, miny := A_ScreenHeight
      Loop, Parse, CtrlHWND, |                           ;go through all controls in selection
        {
          ControlGetPos, x,y,w,h,, ahk_id %A_LoopField%  ;get control pos/size
          maxx := maxx > x + w ? maxx : x + w            ;calculate surrounding box coordinates
          maxy := maxy > y + h ? maxy : y + h
          minx := minx < x ? minx : x
          miny := miny < y ? miny : y
        }
      sW := maxx - minx , sH := maxy - miny              ;calculate size of surrounding box
      sX := minx + ux , sY := miny + uy                        ;calculate global selector coordinates
   }Else{
     ControlGetPos, sx,sy,sw,sh,, % "ahk_id " Selector_vCtrlHWND()  ;get controls pos/size
     sX += ux, sY += uy
   }
  Selector_Set2PosSize(sX, sY, sW, sH)                 ;set new pos/size for selector
  Selector_vPosSize(sX, sY, sW, sH)                    ;store new pos/size
}

;snap selector position to grid
Selector_Set2Grid(){
  GuiHWND := Selector_vGuiHWND()
  WinGetPos, ux, uy, , , ahk_id %GuiHWND%                  ;get pos of user gui
  WinGetPos, sx, sy ,sw, sh, % "ahk_id " Selector_vHWND()  ;get current pos/size of selector

  SysGet_GuiFrameSizes(Gui(GuiHWND,"Toolwindow"), XFrame, YFrame, HCaption)

  rx := sx - ux - XFrame                                   ;calculate relative position
  ry := sy - uy - YFrame - HCaption                                 
  GRID := Grid(GuiHWND)                                            ;get GRID size
  dx := Mod(rx, GRID), dy := Mod(ry, GRID)                         ;calculate delta x/y for snap to grid
  HalfGrid := Grid / 2                                             ;snap to closes grid
  dx -= (dx > HalfGrid ? GRID : 0), dy -= (dy > HalfGrid ? GRID : 0)                    
  Selector_Set2PosSize(sX - dx, sY - dy, sW, sH)                      ;set selector to new position
}

;move selector by increments with hotkeys up/down/left/right
Selector_SetMoveIncrement:
  If Selector_vCtrlHWND(){
    Selector_Size(False)
    Selector_SetMoveIncrement(A_ThisHotkey)
    Selector_MoveCtrlIncrement()
    Grid_Draw(Selector_vGuiHWND())                                ;redraw grid
    ToolTip_ShowTips4Pos()
    Sleep, 20
    Selector_Size(True)
  }
Return

;position Selector by an increment per hotkey press
Selector_SetMoveIncrement(Direction){
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_vHWND()          ;get current pos/size of selector
  Selector_vPosSize(sx, sy, sw, sh)                             ;store old position 
  If Direction = Up                                             ;increment position
      sy-- 
  Else If Direction = ^Up
      sy--, sh++
  Else If Direction = +^Up
      sh--
  Else If Direction = Down
      sy++
  Else If Direction = ^Down
      sh++
  Else If Direction = +^Down
      sy++, sh--
  Else If Direction = Left
      sx--
  Else If Direction = ^Left
      sx--, sw++
  Else If Direction = +^Left
      sw--
  Else If Direction = Right
      sx++
  Else If Direction = ^Right
      sw++
  Else If Direction = +^Right
      sx++, sw--
  Selector_Set2PosSize(sX, sY, sW, sH)                          ;position selector
}

;###############################################################################
;##################   Draw a frame for multi selection                                                       
;###############################################################################

;drag a frame for selection
;called with timer when:
; - user clicked with LButton on gui without selecting a control
Selector_DragFrame:
  Selector_DragFrameDraw()
  If !GetKeyState("LButton")        ;end of drag on release of LButton
    Selector_DragFrameEnd()
Return
;draw a frame between start and end coords 
Selector_DragFrameDraw(X="", Y="", Action=""){
  static sX, sY
  If X                                            ;initialized start coordinates
    Return sX := X, sY := Y
  MouseGetPos, eX, eY                             ;get current mouse pos
  X := sX > eX ? eX : sX, Y := sY > eY ? eY : sY  ;calculate frame position
  W := Abs(sX - eX), H := Abs(sY - eY)            ;calculate size
  Selector_Set2PosSize(X, Y, W, H)                   ;show frame
  If Action = End                                 ;for last call store new position
    Selector_vPosSize(x,y,w,h)
}
;end of frame drag (when user releases LButton)
Selector_DragFrameEnd(){
  SetTimer, Selector_DragFrame, Off
  Selector_DragFrameDraw("", "", "End")
  Selector_SelectCtrlsByFramePosSize()   ;select controls based on frame pos/size
  If Selector_vCtrlHWND() {              ;update selector to new selection
    Selector_Set2Ctrl()
    ToolTip_ShowTips4Pos()
  }Else
    Selector_Set2PosSize(-10, -10)
  Messages_SetHandlers( True )
  Grid_Draw(Selector_vGuiHWND())
  Sleep, 20
  Selector_Size(True)
}

;select controls based on frame pos/size
; called when:
;   - frame was dragged to select controls
Selector_SelectCtrlsByFramePosSize(){
  GuiHWND := Selector_vGuiHWND()                          ;which Gui
  WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()    ;get pos of user gui
  Selector_vPosSize(BoxX, BoxY, BoxW, BoxH)               ;get frame dimentions
  BoxX -= ux, BoxY -= uy                                  ;relative position
  BoxW += BoxX, BoxH += BoxY
  ListOfCtrlHWND := Gui_GetAllCtrlHWNDWTabActive(GuiHWND) ;get only controls visible on tab pages and gui
  Loop, Parse, ListOfCtrlHWND, |                         ;go through all controls on gui
    {                                           
      ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%     ;get current controls pos
      If ( cx >= BoxX AND cy >= BoxY AND cx + cw <= BoxW AND cy + ch <= BoxH ){
        ControlsInBox .= "|" . A_LoopField
        If ("Tab" = Control(A_LoopField,"Type")){           ;when a tab is in selection
          ListOfCtrlOnTab := Tab_GetAllCtrlHWND(A_LoopField)          ;get all controls of that tab
          Loop, Parse, ListOfCtrlOnTab, |
          {                                                        ;and check if they are in box too
             ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%
             If ( cx >= BoxX AND cy >= BoxY AND cx + cw <= BoxW AND cy + ch <= BoxH )
                ControlsInBox .= "|" . A_LoopField
          }
        }
      }
    }
  StringTrimLeft, ControlsInBox, ControlsInBox, 1
  Sort, ControlsInBox, U D|
  If ControlsInBox
      Selector_vCtrlHWND(ControlsInBox)                  ;set selection
  Else
      Return 0
}

;###############################################################################
;##################   Resize selector                                                       
;###############################################################################

;selector gui gets resized
Selector_Size:
  Selector_Size()
Return
Selector_Size(NewState=""){
  static State := True
  If NewState is not space    ;when called with state
    Return State := NewState         
  If !State
      Return
    
  If InStr(CtrlHWND := Selector_vCtrlHWND(),"|") {             ;when multi selection
    SetTimer, Selector_Multi2SelectorAfterSize, 10 ;update controls according to new selector pos/size when resize finished
    Settimer, ToolTip_ShowTips4Pos, -50
    Selector_UpdateAnchorAfterSize(Selector_vGuiHWND(),CtrlHWND)
  }Else If CtrlHWND{                ;when single selection
    If (GRID(Selector_vGuiHWND()) AND !GetKeyState("LControl")){  ;when Grid and NOT LCtrl down
      State := False                                     ;disable itself to resize selector, without calling itself
      Selector_MoveCtrl2SelectorWGrid()                  ;update control in grid
      SetTimer, Selector_UpdateSet2PosAfterSize, 50      ;update selector to new control position when resize finished
      State := True                                      ;enable itself
    }
    else Selector_MoveCtrl2Selector()              ;otherwise move the control directly to follow the selector
    Settimer, ToolTip_ShowTips4Pos, -50
    Selector_UpdateAnchorAfterSize(Selector_vGuiHWND(),CtrlHWND)
  }Else {
    State := False                                     ;disable itself to resize selector, without calling itself
    Selector_Set2PosSize(-10,-10)
    ToolTip_ShowTips4Pos("Hide")
    State := True                                      ;enable itself
  }
  Grid_Draw(Selector_vGuiHWND())
}

Selector_UpdateAnchorAfterSize(gHWND,cHWND){
  static GuiHWND, CtrlHWND
  GuiHWND := gHWND
  CtrlHWND := cHWND
  SetTimer, UpdateAnchorAfterSize, -500
  Return

  UpdateAnchorAfterSize:
    If (AnchoredCtrls := Gui(GuiHWND, "_AnchoredCtrls"))
      Loop, Parse, AnchoredCtrls, |
        If InStr(CtrlHWND, A_LoopField)
          Control_Anchor(GuiHWND, A_LoopField, "")
  Return
}

;update selector to control dimentions 
; called when: 
;   - after selector resize with GRID in single selection mode
;   - after selector move with GRID in single selection mode (Messages_OnMouseMove())
Selector_UpdateSet2PosAfterSize:
  If !GetKeyState("LButton"){
      SetTimer, Selector_UpdateSet2PosAfterSize, Off
      Selector_Set2Ctrl()                    ;update selector to control dimentions
      Grid_Draw(Selector_vGuiHWND())         ;Redraw grid
  }
Return

;update controls according to new selector pos/size when resize finished
Selector_Multi2SelectorAfterSize:
  If !GetKeyState("LButton"){                           ;wait for LButton release, then resize finsihed
      SetTimer, Selector_Multi2SelectorAfterSize, Off        ;disable timer 
      Selector_MoveCtrl2Selector()                           ;update controls to new selector position/size
      Grid_Draw(Selector_vGuiHWND())                         ;redraw grid 
  }
Return

;###############################################################################
;##################   Update control pos/size according to selectors pos/size                                                       
;###############################################################################

; Change control pos/size by increments of selector's size and position
; called when:
;   - after selector is incremented with hotkeys in multi selection
Selector_MoveCtrlIncrement(){
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_vHWND()        ;get current pos/size of selector
  CtrlHWND := Selector_vCtrlHWND()
  Selector_vPosSize(dx, dy, dw, dh)                           ;get selectors previous pos/size
  dx -= sx, dy -= sy, dw -= sw, dh -= sh
  Loop, Parse, CtrlHWND, |                                    ;go through all controls in selection
    {
      ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%                   ;get current controls pos
      ControlMove,, cx - dx, cy - dy, cw - dw, ch - dh, ahk_id %A_LoopField%  ;move control to new pos
    }
  Selector_vPosSize(sx,sy,sw,sh)                              ;store new selector pos/size
  SetTimer, Gui_Refresh, -100                                 ;refresh gui after a timeout
}

; Update control pos/size to match selector's size and position
; called when:
;   - selector resizes in single selection
;   - after selector resize in multi selection
;   - after mouse move with LButton down on selector  
Selector_MoveCtrl2Selector:
  Selector_MoveCtrl2Selector()
Return
Selector_MoveCtrl2Selector(){
  WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()        ;get pos of user gui
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_vHWND()        ;get current pos/size of selector
  If InStr(CtrlHWND := Selector_vCtrlHWND(),"|"){             ;when multi selection
    Selector_vPosSize(dx, dy, dw, dh)                                ;get selectors previous pos/size
    fx := sw / dw , fy := sh / dh                                    ;calculate factor for x/y
    dx := sx - dx + (fx - 1) * (ux - dx) , dy := sy - dy + (fy - 1) * (uy - dy) ;calculate offset for x/y

    ;changing controls position and size based on change in selector width/height doesn't work too well => truncation error
    Loop, Parse, CtrlHWND, |                                 ;go through all controls in selection
      {
        ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%                   ;get current controls pos
        ControlMove,, cx * fx + dx, cy * fy + dy, cw * fx, ch * fy, ahk_id %A_LoopField%  ;move control to new pos
      }
    Selector_vPosSize(sx,sy,sw,sh)                               ;store new selector pos/size
  }Else If CtrlHWND{                                          ;when Single selection
    ControlMove,, sx-ux, sy-uy, sw, sh, ahk_id %CtrlHWND%            ;move control to selector pos/size
  }
  SetTimer, Gui_Refresh, -100                                 ;refresh gui after a timeout
}

; Update single control pos/size to match selector's size and position when GRID is used
; called when:
;   - selector resizes
;   - after mouse move with LButton down on selector with "PositionOnly"
Selector_MoveCtrl2SelectorWGrid(Type="") {
  GuiHWND := Selector_vGuiHWND()
  WinGetPos, ux, uy, , , ahk_id %GuiHWND%                  ;get pos of user gui
  WinGetPos, cx, cy ,cw, ch, % "ahk_id " Selector_vHWND()       ;get current pos/size of selector

  SysGet_GuiFrameSizes(Gui(GuiHWND,"Toolwindow"), XFrame, YFrame, HCaption)

  cx -= ux, cy -= uy                                      ;calculate relative position
  GRID := Grid(GuiHWND)                                   ;get GRID size
  dx := Mod(cx - XFrame, GRID)                            ;calculate x/y with snap to grid
  dy := Mod(cy - YFrame - HCaption, GRID)             
  dx -= (dx ? GRID : 0),cx -= dx, dy -= (dy ? GRID : 0),cy -= dy
  If !(Type = "PositionOnly"){                                    ;when also w/h should be adjusted (on selector resize)
    dw := Mod(cw, GRID), dh := Mod(ch, GRID)                      ;calculate w/h with snap to grid 
    cw -= (cw > dw ? dw : 0), ch -= (ch > dh ? dh : 0)
  }
  ControlMove,,cx,cy,cw,ch, % "ahk_id " Selector_vCtrlHWND()      ;move control to new pos
}

;###############################################################################
;##################   Delete control(s) in selection                                                       
;###############################################################################

;context sensitive hotkey to delete selection got pressed
Selector_DeleteControl:
  Control_DeleteControlList(Selector_vCtrlHWND())   ;delete selected control(s)
  Selector_vCtrlHWND("No")                          ;clear selection
  Selector_Set2PosSize(-10, -10)                    ;hide selector
  ToolTip_ShowTips4Pos("Hide")
  Grid_Draw(Selector_vGuiHWND())                    ;redraw grid when control(s) deleted
Return
