;Title:		Properties GUI
;			Settings/Properties editor/viewer.
;
;Version:	0.2

;------------------------------------------------------------------------------------------
; Function:		Clear
;				Clear Property window
;
Properties_Clear(){
	global
	loop, % SS_GetRowCount(Properties_hSheet)
			SS_DeleteRow(Properties_hSheet, 1)
}

;------------------------------------------------------------------------------------------
; Function:		GetValue
;				Get property value
;
; Parameters:
;				pName		- Proeprty name for which to get value
;
; Returns:
;				Value
Properties_GetValue( pName ) {
	local row := Properties_Find( pName ), tpe = SS_GetCellType( Properties_hSheet, 2, row ) 

	if pName = 
		return 

	if tpe in 11,12,6			;CheckBox,ComboBox,Integer
			return SS_GetCellData( Properties_hSheet, 2, row) 
	else	return SS_GetCellString( Properties_hSheet, 2, row)
}

;------------------------------------------------------------------------------------------
; Function:		Add
;				Add property to the list on given postion
;
; Parameters:
;				pProperty	- Property defintion - multiline string containing 1 property attribute
;							  per line. 
;				pNo			- Property position. -1 (default) means property will be appended
;
; Atributes:
;				Name		- Name of the property to be displayed in first column
;				Type		- Type of the property. Currently supported types are:
;							  Text, Button, WideButton, CheckBox, ComboBox, Integer, Hyperlink, Separator
;				Text		- Default text. For ComboBox item this contains pipe delimited list of 
;							  items to be added to the ComboBox. For CheckBox 1 or 0
;
Properties_Add( pProperty, pNo=-1 ){
	
	local j, desc, hCtrl := Properties_hSheet
	local Name, Type, Separator, Text
	local PB := Properties_PB,  PF := Properties_PF  ,VF := Propeties_VF,	VB := Properties_VB

	local state:="Default", fnt1=0, fnt2=1, txtal:="RIGHT MIDDLE", tpe, imgal="MIDDLE RIGHT", txtal2="MIDDLE LEFT"

 ;parse property into local variables
	loop, parse, pProperty, `n
	{
		j := InStr(A_LoopField, "=")
		desc   := SubStr(A_LoopField, 1, j-1)
		%desc% := SubStr(A_LoopField, j+1, StrLen(A_LoopField))
	}

 ;insert row
	pNo := pNo=-1 ? SS_GetRowCount(hCtrl)+1 : pNo
	SS_InsertRow(hCtrl, pNo)		

 ;set defaults
	tpe := Type 
	if (Type="Separator") {
		tpe:= "ForceText", state := "Locked", fnt1 := 2, Name := Text , Text := ""
		VB := PB := Properties_SB, VF := PF := Properperties_SF
		txtal := "LEFT MIDDLE"

		SS_ExpandCell( hCtrl, 1, pNo, 2, pNo )
	}

	if (Type="HyperLink")
		state := "LOCKED", fnt2 := 3

	if (type="ComboBox")
		 Text := SS_CreateCombo(hCtrl, Text)
	
	if (Type="Button")		
		 tpe := "Button Text ForceType"

	if (Type="WideButton")	
		 tpe := "WideButton Text ForceType", status="LOCKED", txtal2="CENTER MIDDLE"
	
	if (type="CheckBox")
		 tpe := "CheckBox"
	

 ;set row
	SS_SetCell(	hCtrl, 1, pNo
				,"type=FORCETEXT", "txt=" Name
				,"bg=" PB, "fg=" PF
				,"state=LOCKED", "txtal=" txtal, "fnt=" fnt1, "h=" Properties_rowH)

	SS_SetCell( hCtrl, 2, pNo
				,"type=" Tpe 
				,type="CheckBox" ? "data=" Text : "txt=" Text
				,"bg=" VB, "fg=" VF
				,"txtal=" txtal2, "imgal=" imgal
				,"fnt=" fnt2, "state=" state)

	;refresh (happens when you set focus)	
	ControlFocus,, ahk_id %hCtrl%
}


;------------------------------------------------------------------------------------------
; Function:		AddFromText
;				Add properties from file
;
; Parameters:
;				pText	- Text from which to import properties. The text contains N property definitions
;						  started by "[Property]" keyword. 
;
Properties_AddFromText( pText ) {
 ;remove blank lines
	pText := RegExReplace(pText, "\R\s*(?=\R)")
	
 ;remove white space before & after section
	pText := RegExReplace(pText, "[ \t]*(\[.+?\])[ \t]*(?=\R)", "$1")

 ;make it more convenient for loop, parse
	StringReplace, pText, pText, `r`n, `n, A

	re := "\Q[Property]\E\n.+?(?=(\R\Q[Property]\E)|\R*$)"
	loop {
		if (pText ="")
			break

	  ;get next property
		RegExMatch(pText, re, cp)
		pText := RegExReplace(pText, re, "", "", 1)
		StringTrimLeft, pText, pText, 1
		StringReplace, cp, cp, [Property]`n, 	  

		Properties_Add(cp)
	}
}


;------------------------------------------------------------------------------------------
; Function:		AddFromFile
;				Add properties from file
;
; Parameters:
;				pFile	- File from which to import properties. The file contains N property definitions
;						  started by "[Property]" keyword. 
;
Properties_AddFromFile( file ) {
	FileRead, pText, %file%
	Properties_AddFromText( pText )
}

;------------------------------------------------------------------------------------------
; Function:		Show
;				Show property window. Sets Properties_visible to true
;
; Parameters:
;				x, y, w, h	- Optional parameters for position and size
;       x           - If x is set to "Hide", property window hides
;
Properties_Show(x="", y="",w="", h="") {
	local sp
	static init=0, Properties_visible:=False
	
	If x = Hide
	{
  	Properties_visible := false
  	Gui, %Properties_GUI%:Hide
    Return
  }
  
  If Properties_visible
	   Return

	if !init++
		h := Properties_height, w := Properties_width 

	sp .= h !="" ? " h" h :
	sp .= w !="" ? " w" w :
	sp .= x !="" ? " x" x :
	sp .= y !="" ? " y" y :
	
	Gui, %Properties_GUI%:Show, %sp% NA, %Properties_title%
	Properties_visible := true

;	Properties_Size()
}

;------------------------------------------------------------------------------------------
; Function:		Create
;				Creates property window.
;
; Parameters:
;				pHandler	- Notification handler. PTY_NAME & PTY_EVENT global variables contain event details
;				title		- GUI title
;				flags		- Space separated creation flags
;				GuiNum		- Optional GUI num, By default 89
;
; Flags:		
;				ToolWindow	- Makes PGUI tool like.
;				Resize		- Makse PGUI resizable
;				NoCaption	- Removes PGUI caption
;				NoCombo		- Don't create ComboBox
;				H[num]		- Row Height
;
; Notifications:
;				PTY_NAME	- Name of the property that generated event
;				PTY_EVENT	- Can be one of the following: BeforeUpdate, AfterUpdate, Button, Hyperlink
;
; Notes:		
;				Properties_hCombo contains handle to the combo box
;
Properties_Create(pHandler, title="", flags="", GuiNum=89) {
	local header = 20, height=400, width = 250, col1W=110
	local col2W, hCtrl, guiOpt, noCombo, rowH, status

	col2W := width - col1W - 4

 ;parse flags
	guiOpt .= InStr(flags, "ToolWindow") ? " +ToolWindow" :
	guiOpt .= InStr(flags, "Resize")	 ? " +Resize" :
	guiOpt .= InStr(flags, "NoCaption")  ? " -Caption" :

	noCombo := InStr(flags, "NoCombo")

	RegExMatch(flags, "(?<=[H|h])(\d{1,2})(?=([ ]|$))", rowH) 
    if ( rowH = "") 
		 rowH := 20 
  	
	Gui, %GuiNum%:+LastFound %guiOpt% +LabelProperties_ +MinSize200x%header%

  ;set globals
	Properties_title	:= title
	Properties_GUI		:= GuiNum
	Properties_height	:= height
	Properties_width	:= width
	Properties_hwnd		:= WinExist()
	Properties_handler  := pHandler
	Properties_rowH		:= rowH

	if !noCombo	
			Gui, %GuiNum%:Add, ComboBox, x0 y0 w%width% h20 HWNDProperties_hCombo  vProperties_Combo
	else	header := 0

	hCtrl := Properties_hSheet := SS_Add( Properties_hwnd, 0, header, width+20, height - header, "STATUS VSCROLL GRIDMODE CELLEDIT COLSIZE ROWSELECT")
	SS_SetRowHeight(hCtrl, 0, 5)

	SS_SetColWidth(hCtrl, 1, col1W)
	SS_SetColWidth(hCtrl, 2, col2W)
	SS_SetColCount(hCtrl, 2)
	SS_SetRowCount(hCtrl, 0)

	Properties_SetFont("Property",	"s11        ,MS Sans Serif")
	Properties_SetFont("Value",		"s12        ,Courier New")	
	Properties_SetFont("Separator", "s10 italic ,Arial")	
	Properties_SetFont("Hyperlink",	"s08 italic ,Arial")

	SS_SetHandler("Properties_OnNotify")	

	Hotkey, ifWinActive, ahk_id %Properties_hwnd%
	Hotkey, ESC, Properties_Escape
	Hotkey, ifWinActive

}

;------------------------------------------------------------------------------------------
; Function:		SetColors
;				Set colors for any subset of property elements
;
; Parameters:
;				colors	- String containing space separated colors for property elements
;
; Colors:		
;				PB PF - property bg & fg 
;				VB VF - value bg & fg
;				SB SF - separator bg & fg
; Example:
;>				Properties_SetColors("pbAAEEAA sbBFFFFF")  ;set property and separator background color  
;
Properties_SetColors(colors){
	local token, c

	Properties_PB := Properties_VB := Properties_SB := 0xFFFFFF
	Properties_PF := Properties_VF := Properties_SF := 0x0
	Loop, Parse, colors, %A_Space%
	{
		if A_LoopField =
			continue

		StringLeft c, A_LoopField, 2
		StringTrimLeft token, A_LoopField, 2
		
		Properties_%c% := "0x" token
	}
}

;------------------------------------------------------------------------------------------
; Function:		SetFont
;				Set font for 1 Propety element
;
; Parameters:
;				element	- 1 of the 3 available elements: Property, Value, Separator
;				font	- Font description in AHK format
;

Properties_SetFont(element, font) {
	local idx
	
	if (element="Property")
		idx := 0
	if (element="Value")
		idx := 1
	if (element="Separator")
		idx := 2
	if (element="HyperLink")
		idx := 3

	SS_SetFont(Properties_hSheet, idx, font)	;properties
}

;------------------------------------------------------------------------------------------
; Sets column width of "Value" column to match gui width - column1
Properties_size(){
	local cw, pw, sRect
	
	VarSetCapacity(sRect, 16)
	DllCall("GetClientRect", "uint", Properties_hwnd, "uint", &sRect) 	
	pw := ExtractInteger(sRect, 8)-2
	cw := SS_GetColWidth(Properties_hSheet, 1)

	SS_SetColWidth(Properties_hSheet, 2, pw-cw)
}

Properties_size:
	Properties_Anchor("SPREAD_SHEET1", "wh")
	Properties_Anchor("Properties_Combo", "w")
	Properties_Size()
return

;------------------------------------------------------------------------------------------

Properties_escape:
Properties_close:
	Properties_Show("Hide")
return

;------------------------------------------------------------------------------------------
; Returns row of the given property or separator
Properties_find( pName ) {
	local cnt

	cnt := SS_GetRowCount(Properties_hSheet)
	loop, %cnt%
		if SS_GetCellString(Properties_hSheet, 1, A_Index) = pName
			return A_Index
	
	return 0
}

;------------------------------------------------------------------------------------------
; Notification handler
Properties_onNotify(wparam, lparam, msg, nhwnd){
	local hwnd, idCtrl, code,		name, col
	static SPRN_SELCHANGE=1, SPRN_BEFOREEDIT=2, SPRN_AFTEREDIT=3, SPRN_BEFOREUPDATE=4, SPRN_AFTERUPDATE=5, SPRN_HYPERLINKENTER=6, SPRN_HYPERLINKLEAVE=7, SPRN_HYPERLINKCLICK=8, SPRN_BUTTONCLICK=9
	
	hwnd   :=  DecodeInteger(lparam)
    idCtrl :=  DecodeInteger(lparam+4)
    code   :=  DecodeInteger(lparam+8)

;	OutputDebug %code%
;	return

	if (code = SPRN_SELCHANGE) {
		col  := DecodeInteger(lparam+16)
		Properties_srow := DecodeInteger(lparam+20)

		SS_GetCell(Properties_hSheet, 2, Properties_srow, "Properties_i")
		Properties_stype := Properties_i_hType

  		;EncodeInteger(lparam+24, 1)		;fcancel (prevent selection)
		if col = 1						
			SetTimer, Properties_OnSheetSelect, -10				;if user selected first column, swich to second

		return
	}

	
	PTY_NAME  := SS_GetCellString(Properties_hSheet, 1, Properties_srow)
	if (code = SPRN_BEFOREEDIT){
		PTY_EVENT := "BeforeEdit"
		gosub %Properties_Handler%
		return
	}

	if (code = SPRN_AFTEREDIT){
		PTY_EVENT := "AfterEdit"
		gosub %Properties_Handler%
		return
	}

	if (code = SPRN_BEFOREUPDATE){
		PTY_EVENT := "BeforeUpdate"
		gosub %Properties_Handler%
		return
	}

	if (code = SPRN_AFTERUPDATE){
		PTY_EVENT := "AfterUpdate"
		gosub %Properties_Handler%
		return
	}

	if (code = SPRN_BUTTONCLICK){
 		PTY_EVENT := "Button"
		gosub %Properties_Handler%
		return
	}

	if (code = SPRN_HYPERLINKCLICK){
 		PTY_EVENT := "Hyperlink"
		gosub %Properties_Handler%
		return
	}
}

Properties_onSheetSelect:
	SS_SetCurrentCell(Properties_hSheet, 2, SS_GetCurrentRow(Properties_hSheet)) 
return

Properties_Anchor(c, a = "", r = false) { ; v3.6 - Titan
	static d
	GuiControlGet, p, Pos, %c%
	If ex := ErrorLevel {
		Gui, %A_Gui%:+LastFound
		ControlGetPos, px, py, pw, ph, %c%
	}
	If !(A_Gui or px) and a
		Return
	i = x.w.y.h./.7.%A_GuiWidth%.%A_GuiHeight%.`n%A_Gui%:%c%=
	StringSplit, i, i, .
	d := a ? d . ((n := !InStr(d, i9)) ? i9 : "")
		: RegExReplace(d, "\n\d+:" . c . "=[\-\.\d\/]+")
	Loop, 4
		x := A_Index, j := i%x%, i6 += x = 3
		, k := !RegExMatch(a, j . "([\d.]+)", v) + (v1 ? v1 : 0)
		, e := p%j% - i%i6% * k, d .= n ? e . i5 : ""
		, RegExMatch(d, "\Q" . i9 . "\E(?:([\d.\-]+)/){" . x . "}", v)
		, l .= p%j% := InStr(a, j) ? (ex ? "" : j) . v1 + i%i6% * k : ""
	If r
		rx = Draw
	If ex
		ControlMove, %c%, px, py, pw, ph
	Else GuiControl, Move%rx%, %c%, %l%
}
